package industries.doublehelix.pwste;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.TreeMap;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;

import com.github.luben.zstd.Zstd;

import net.jpountz.lz4.LZ4Factory;
import net.jpountz.lz4.LZ4FastDecompressor;

/**
 * Class to hold one timelapse (all images)
 */
public class TimeLapse {
	private Path sourcePath;
	private String rawName;
	private String processedName;
	private List<byte[]> images = new ArrayList<>();
	private boolean isCompleteSave;
	
	public TimeLapse(Path sourceDir, boolean isCompleteSave) {
		sourcePath = sourceDir;
		this.isCompleteSave = isCompleteSave;
		
		this.rawName = sourceDir.getFileName().toString();
		if ((!isCompleteSave && !this.rawName.toLowerCase().endsWith("timelapse")) || (isCompleteSave && !this.rawName.toLowerCase().endsWith("timelapse-comp"))) {
			throw new IllegalArgumentException("Not a timelapse save folder");
		}
		
		String nameParts[] = this.rawName.split("_");
		if (nameParts.length < 3) throw new IllegalArgumentException("Not a timelapse save folder");
		StringBuilder processedNameBuilder = new StringBuilder(StringUtils.capitalize(nameParts[0].toLowerCase()) + ":");
		// last one is always TIMELAPSE if in progress or TIMELAPSE-COMP is completed
		if (!isCompleteSave) processedNameBuilder.append("[In Progress] ");
		for (int i = 1; i < nameParts.length - 1; ++i) {
			if (i == 2) processedNameBuilder.append(" -"); // let's see if this'll hold up
			processedNameBuilder.append(" ").append(StringUtils.capitalize(nameParts[i].toLowerCase()));
		}
		this.processedName = processedNameBuilder.toString();
		//System.out.println("procname: " + processedName);
	}
	
	public synchronized void load() throws IOException {
		if (!images.isEmpty()) return; // already loaded
		if (isCompleteSave) {
			loadComplete();
		} else {
			loadInProgress();
		}
	}
	
	private synchronized void loadInProgress() throws IOException {
		// in progress saves are individual save files with foldername_number.sav
		// each save file contains a single lz4 compressed image (see getImageFromInprogressSaveFile for file structure details)
		
		TreeMap<Integer, Path> sortedFiles = new TreeMap<>();
		try (Stream<Path> filePathStream = Files.list(this.sourcePath)) {
			// the files are already numbered.. sadly not in a sortable way without some extra work or we'd get 0, 1, 100, 11
			filePathStream.forEach(filePath -> {
				String fileName = filePath.getFileName().toString();
				if (Files.isRegularFile(filePath) && fileName.toLowerCase().endsWith(".sav")) {
					String[] filenameParts = fileName.split("_");
					String numberStr = filenameParts[filenameParts.length - 1];
					// this may cause IndexOutOfBounds or NumberFormatException, but that means there's a problem with this 'lapse anyway
					numberStr = numberStr.substring(0, numberStr.length() - 4);
					sortedFiles.put(Integer.parseInt(numberStr), filePath);
				}
			});
		}
		for (int key : sortedFiles.keySet()) {
			Path saveFile = sortedFiles.get(key);
			images.add(getImageFromInprogressSaveFile(saveFile));
		}
	}
	
	private byte[] getImageFromInprogressSaveFile(Path saveFile) throws IOException {
		// file structure:
		// all numbers are little endian
		// 4 bytes width
		// 4 bytes height
		// 4 bytes uncompressed image size
		// 4 bytes compressed image size
		// 4 bytes uncompressed image size (again)
		// LZ4 compressed image data
		// afterwards: don't care
		
		byte[] file = Files.readAllBytes(saveFile);

		ByteBuffer uncompressedSizeWrap = ByteBuffer.wrap(Arrays.copyOfRange(file, 16, 20));
		uncompressedSizeWrap.order(ByteOrder.LITTLE_ENDIAN);
		int uncompressedSize = uncompressedSizeWrap.getInt();
		
		LZ4Factory factory = LZ4Factory.fastestInstance();
		LZ4FastDecompressor decompressor = factory.fastDecompressor();
		byte[] restored = new byte[uncompressedSize];
		decompressor.decompress(file, 20, restored, 0, uncompressedSize);
		
		return restored;
	}
	
	private synchronized void loadComplete() throws IOException {
		// complete saves are made up of three save files.
		// foldername_0.sav and foldername_100.sav for the first and last pic, and foldername_0-100.sav containing all the pics of the timelapse (foldername without the -COMP)
		// of interest is only 0-100.sav for us
		Path completeSave = null;
		
		List<Path> files;
		try (Stream<Path> filePathStream = Files.list(this.sourcePath)) {
			files = filePathStream.filter(file -> Files.isRegularFile(file) && file.getFileName().toString().toLowerCase().endsWith("0-100.sav")).collect(Collectors.toList());
		}
		if (files == null || files.size() != 1) {
			System.err.println("Coud not find complete save in " + this.sourcePath);
			return;
		}
		
		byte[] file = Files.readAllBytes(files.get(0));
		
		// file structure:
		// all numbers are little endian
		// 4 bytes unknown, perhaps checksum
		// 4 bytes width
		// 4 bytes height
		// 8 bytes uncompressed size of image array
		// 8 bytes compressed size of image array
		// Zstandard compressed image list (of the format of the in progress save files)
		ByteBuffer byteBuffer = ByteBuffer.wrap(file, 12, 8); // start at uncompressed size
		byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
		int uncompressed = (int)byteBuffer.getLong(); // cast to int should be save, no way this is > 1GB, which is the max in a signed int
		
		byte[] zuncompressed = new byte[uncompressed];
		Zstd.decompressByteArray(zuncompressed, 0, uncompressed, file, 28, file.length-28);
		
		// zuncompressed now contains saves in the structure of getImageFromInprogressSaveFile() one after another:
		// all numbers are little endian
		// 4 bytes width
		// 4 bytes height
		// 4 bytes uncompressed image size
		// 4 bytes compressed image size
		// 4 bytes uncompressed image size (again)
		// LZ4 compressed image data
		LZ4Factory factory = LZ4Factory.fastestInstance();
		LZ4FastDecompressor decompressor = factory.fastDecompressor();
		ByteBuffer unzipwrap = ByteBuffer.wrap(zuncompressed);
		unzipwrap.order(ByteOrder.LITTLE_ENDIAN);
		int offset = 0;
		for (int i = 0;; ++i) {
			unzipwrap.position(offset + 8);
			int uncompressedSize = unzipwrap.getInt();
			int compressedSize = unzipwrap.getInt();
			byte[] restored = new byte[uncompressedSize];
			decompressor.decompress(zuncompressed, 20+offset, restored, 0, uncompressedSize);
			
			// the first image is always the completed image (same as last image) for some reason, so skip it
			if (i != 0) {
				images.add(restored);
			}
			
			offset = offset + 20 + compressedSize;
			if (offset >= zuncompressed.length) break;
		}
	}

	public Path getSourcePath() {
		return sourcePath;
	}

	public String getRawName() {
		return rawName;
	}

	public String getProcessedName() {
		return processedName;
	}
	
	@Override
	public String toString() {
		return getProcessedName();
	}

	public List<byte[]> getImages() {
		return Collections.unmodifiableList(this.images);
	}
}
